/* $Id: globus-replica-management.c,v 1.16 2003/03/31 23:04:36 link Exp $ */

/**
 * @file
 *
 * Globus Replica Management Command
 *
 * @see globus_replica_management.h
 */

#include "globus_common.h"
#include "globus_args.h"
#include "globus_replica_management.h"
#include "version.h" /* provides local_version */

#define GLOBUS_REPLICA_MANAGEMENT_BUFSIZE 256

const char * usage_short =
"globus-replica-management [<object>] <action>\n";

const char* usage_long =
"globus-replica-management [<object>] <action>\n"
"globus-replica-management -version[s]\n"
"globus-replica-management -help\n"
"\n"
"    [-manager <manager DN> | -D <manager DN>]\n"
"        directory manager DN for replica catalog authentication,\n"
"        may be specified by the GLOBUS_REPLICA_CATALOG_MANAGER\n"
"        environment variable instead, otherwise prompted for\n"
"    [-password <password file name> | -P <password file name]\n"
"        file containing the password for replica catalog authentication,\n"
"        otherwise prompted for\n"
"\n"
"    -collection <collection url> | -c <collection url>\n"
"        act on the collection in the given replica catalog LDAP URL,\n"
"        of the form ldap://host[:port]/dn, and may be specified via the\n"
"        GLOBUS_REPLICA_CATALOG_COLLECTION environment variable instead\n"
"\n"
"        -create            creates collection information\n"
"        -delete            deletes collection information, successful\n"
"                           if no associated location information \n"
"\n"
"    -location <location name> | -l <location name>\n"
"        act on the given location name, a collection that contains this\n"
"        location must also be specified\n"
"\n"
"        -create <root URL> creates location information with the given\n"
"                           URL of a \"root\" directory in a storage\n"
"                           system, where the protocol may be one of\n"
"                           \"ftp:\", or \"gsiftp:\"\n"
"        -delete            deletes location information\n"
"\n"
"    -files <input file name> | -f <input file name>\n"
"	act on the list of logical file names (a name to a line) in the\n"
"	given input file; for \"-publish\", act on a list of source file URLs\n"
"	with corresponding logical file names (a name and a URL to a line,\n"
"	separated by a space); a collection and a location that\n"
"	(potentially) contain these logical file names must also be\n"
"	specified\n"
"\n"
"        -publish           publishes files to a location, copying physical\n"
"                           files specified by URLs and registering logical\n"
"                           file names to the specified location\n"
"        -copy <target location name>\n"
"                           copies corresponding physical files from the\n"
"                           the specified location to the given location\n"
"        -register          registers logical file names in the specified\n"
"                           location, the logical file names must have\n"
"                           corresponding physical files (whose URLs are\n"
"                           constructed by the given logical file names\n"
"                           in the context of the specified location)\n"
"        -delete            deletes logical file names from a location\n"
"        -delete-files      deletes logical file names from a location\n"
"                           and their corresponding physical files\n"
"\n"
"    -list-rollback-info <replica catalog url>\n"
"        lists outstanding rollback and restart information in the given\n"
"        replica catalog URL\n"
"    -rollback <contact>\n"
"        rolls back an operation represented by the given contact string\n"
"    -restart <contact>\n"
"        restarts an operation represented by the given contact string\n"
"\n"
"    -no-third-party[-transfers] | -notp[t] \n"
"         buffers an ftp transfer locally instead of initiating a\n"
"         third-party transfer (which is the default)\n"
"    -parallel[ism] <number of streams> | -p[ar] <number of streams>\n"
"         number of data streams for an ftp transfer\n"
"    -tcp-buffer-size <buffer size> | -tcpbuf <buffer size>\n"
"         tcp buffer size for an ftp transfer\n"
"    -subject <subject> | -s <subject>\n"
"         subject to match for all servers, or just the source server\n"
"         if the destination server subject is given below\n"
"    -dest-subject <subject> | -ds <subject>\n"
"         subject to match for the destination server\n"
"\n"
"  Example\n"
"\n"
"    globus-replica-management \\\n"
"       -collection \"ldap://foo.anl.gov/lc=B04.16.atm,\\\n"
"               rc=PCMDI Replica Catalog,o=Argonne National Laboratory,\\\n"
"               o=Globus,c=US\" \\\n"
"       -location \"sourceLocationName\" \\\n"
"       -files logicalFilenamesInputfile \\\n"
"       -copy \"destLocationName\"\n"
"\n";

enum { arg_manager = 1, arg_password,
	arg_collection, arg_location, arg_files,
	arg_create, arg_delete, arg_delete_files,
	arg_publish, arg_register, arg_copy,
	arg_no_third_party, arg_parallelism, arg_tcp_buffer_size,
	arg_subject, arg_dest_subject,
	arg_list_rollback_info, arg_rollback, arg_restart,
	optc = arg_restart };

static char * arg_manager_aliases[] = { "-manager", "-m", "-D",
	GLOBUS_NULL };
static char * arg_password_aliases[] = { "-password", "-P",
	GLOBUS_NULL };

static char * arg_collection_aliases[] = { "-collection", "-c",
	GLOBUS_NULL };
static char * arg_location_aliases[] = { "-location", "-l",
	GLOBUS_NULL };
static char * arg_files_aliases[] = { "-files", "-f",
	GLOBUS_NULL};

static char * arg_create_aliases[] = { "-create",
	GLOBUS_NULL };
static char * arg_delete_aliases[] = { "-delete",
	GLOBUS_NULL };
static char * arg_delete_files_aliases[] = { "-delete-files",
	GLOBUS_NULL };

static char * arg_register_aliases[] = { "-register",
	GLOBUS_NULL };
static char * arg_publish_aliases[] = { "-publish",
	GLOBUS_NULL };
static char * arg_copy_aliases[] = { "-copy",
	GLOBUS_NULL };

static char * arg_no_third_party_aliases[] = { "-no-third-party",
	"-no-third-party-transfers", "-notp", "-notpt",
	GLOBUS_NULL };
static char * arg_parallelism_aliases[] = { "-parallelism",
	"-p", "-par", "-parallel",
	GLOBUS_NULL };
static char * arg_tcp_buffer_size_aliases[] = { "-tcp-buffer-size",
	"-tcpbuf", "-tcp-bs",
	GLOBUS_NULL };
static char * arg_subject_aliases[] = { "-subject",
	"-s", "-ss",
	GLOBUS_NULL };
static char * arg_dest_subject_aliases[] = { "-dest-subject",
	"-ds",
	GLOBUS_NULL };

static char * arg_list_rollback_info_aliases[] = { "-list-rollback-info",
	GLOBUS_NULL };
static char * arg_rollback_aliases[] = { "-rollback",
	GLOBUS_NULL };
static char * arg_restart_aliases[] = { "-restart",
	GLOBUS_NULL };

static globus_args_option_descriptor_t optv[] = {
	{ arg_manager, arg_manager_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_password, arg_password_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },

	{ arg_collection, arg_collection_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_location, arg_location_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_files, arg_files_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },

	{ arg_create, arg_create_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_delete, arg_delete_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_delete_files, arg_delete_files_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },

	{ arg_register, arg_register_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_publish, arg_publish_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_copy, arg_copy_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },

	{ arg_no_third_party, arg_no_third_party_aliases, 0,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_parallelism, arg_parallelism_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_tcp_buffer_size, arg_tcp_buffer_size_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_subject, arg_subject_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_dest_subject, arg_dest_subject_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },

	{ arg_list_rollback_info, arg_list_rollback_info_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_rollback, arg_rollback_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL },
	{ arg_restart, arg_restart_aliases, 1,
		GLOBUS_NULL, GLOBUS_NULL } };

#define globus_l_replica_management_arg_error(msg) \
{ \
    globus_libc_fprintf( stderr, \
	    "%s\nUse -help to display full usage.\n", (msg) ); \
    globus_module_deactivate_all(); \
    exit(1); \
}

/* specifies an object, with precedence given to files over locations,
 * and locations over collections */
#define globus_l_replica_management_specify_object(arg) \
{ \
    if (object < arg) { \
	object = arg; \
    } \
}

/* checks that no more than one action is specified  */
#define globus_l_replica_management_specify_action(arg) \
{ \
    if (action != 0) { \
	globus_l_replica_management_arg_error( \
	    "Specify no more than one action." ); \
    } else { \
	action = arg; \
    } \
}

#define globus_l_replica_management_check_error(err) \
{ \
    if (err != GLOBUS_SUCCESS) \
    { \
	globus_libc_fprintf( stderr, "%s\n", \
		globus_object_printable_to_string( \
			globus_error_get(err) ) ); \
	errors++; \
    } \
}

void
globus_l_replica_catalog_print_attributes(
	globus_replica_catalog_attribute_set_t attributes )
{
    char * name;
    char ** values;
    int i;
    
    globus_replica_catalog_attribute_set_first(&attributes);
    while (globus_replica_catalog_attribute_set_more(&attributes))
    {
	globus_replica_catalog_attribute_set_get_name(&attributes, &name);
	globus_replica_catalog_attribute_set_get_values(&attributes, &values);
	for (i = 0; values[i]; i++ )
	{
	    globus_libc_printf("%s=%s\n", name, values[i]);
	}
	globus_libc_free(name);
	globus_replica_catalog_attribute_set_free_values(values);
	globus_replica_catalog_attribute_set_next(&attributes);
    }

    return;
}

void
globus_l_replica_catalog_print_entries(
	globus_replica_catalog_entry_set_t entries )
{
    globus_replica_catalog_attribute_set_t attributes;
    char * url;

    globus_replica_catalog_entry_set_first(&entries);
    while (globus_replica_catalog_entry_set_more(&entries))
    {
	globus_replica_catalog_attribute_set_init(&attributes);
	globus_replica_catalog_entry_set_get_attributes( &entries,
		&attributes );
	globus_replica_catalog_entry_set_get_url(&entries, &url);
	globus_libc_printf("contact: %s\n", url);
	globus_libc_free(url);
	globus_l_replica_catalog_print_attributes(attributes);
	globus_libc_printf("\n");
	globus_replica_catalog_attribute_set_destroy(&attributes);
	globus_replica_catalog_entry_set_next(&entries);
    }

    return;
}

int
main(
	int argc,
	char *argv[] )
{
    globus_list_t *			options_found        = GLOBUS_NULL;
    globus_list_t *			list                 = GLOBUS_NULL;
    globus_args_option_instance_t *	option               = GLOBUS_NULL;
    int					object               = 0;
    int					action               = 0;
    char *				collection_url       = GLOBUS_NULL;
    char *				location_name        = GLOBUS_NULL;
    char *				input_file_name      = GLOBUS_NULL;
    char *				root_url             = GLOBUS_NULL;
    char *				target_location_name = GLOBUS_NULL;
    char *				replica_catalog_url  = GLOBUS_NULL;
    char *				contact_string       = GLOBUS_NULL;
    char *				manager              = GLOBUS_NULL;
    char *				password_file_name   = GLOBUS_NULL;
    char *				password             = GLOBUS_NULL;
    char **				logical_file_names   = GLOBUS_NULL;
    char **				source_urls          = GLOBUS_NULL;
    globus_bool_t			third_party_transfer = GLOBUS_TRUE;
    int					num_streams          = 0;
    int					tcp_buffer_size      = 0;
    char * 				subject              = GLOBUS_NULL;
    char * 				dest_subject         = GLOBUS_NULL;
    int					errors               = 0;
    int					i;
    FILE *				fd;
    globus_replica_management_handle_t			handle;
    globus_replica_management_handleattr_t		handleattrs;
    globus_replica_catalog_collection_handleattr_t	collection_attrs;
    globus_result_t *					err;
    globus_replica_catalog_entry_set_t			entries;
    globus_ftp_client_operationattr_t			source_operationattrs;
    globus_ftp_client_operationattr_t			dest_operationattrs;
    globus_ftp_control_parallelism_t			parallelism;
    globus_ftp_control_tcpbuffer_t			tcp_buffer;

    if (globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS)
    {
	globus_libc_fprintf( stderr,
		"Error initializing common module.\n" );
	exit(1);
    }

    if (globus_module_activate(GLOBUS_REPLICA_MANAGEMENT_MODULE)
	    != GLOBUS_SUCCESS )
    {
	globus_libc_fprintf( stderr,
		"Error initializing replica management module.\n" );
	exit(1);
    }

    /*
     * scan arguments
     */

    if ( (i = globus_args_scan( &argc, &argv,
		optc, optv,
		"globus-replica-management-program", &local_version,
		usage_short, usage_long, 
		&options_found, GLOBUS_NULL )) < 0 )
    {
	globus_module_deactivate_all();
	exit(i == GLOBUS_FAILURE ? 1 : 0);
    }

    for ( list = options_found;
	    !globus_list_empty(list);
	    list = globus_list_rest(list) ) {
	option = globus_list_first(list);
	switch (option->id_number)
	{
	case arg_manager:
	    manager = globus_libc_strdup(option->values[0]);
	    break;
	case arg_password:
	    password_file_name = globus_libc_strdup(option->values[0]);
	    break;

	case arg_collection:
	    globus_l_replica_management_specify_object(arg_collection);
	    collection_url = globus_libc_strdup(option->values[0]);
	    break;
	case arg_location:
	    globus_l_replica_management_specify_object(arg_location);
	    location_name = globus_libc_strdup(option->values[0]);
	    break;
	case arg_files:
	    globus_l_replica_management_specify_object(arg_files);
	    input_file_name = globus_libc_strdup(option->values[0]);
	    break;

	case arg_create:
	    globus_l_replica_management_specify_action(arg_create);
	    break;
	case arg_delete:
	    globus_l_replica_management_specify_action(arg_delete);
	    break;
	case arg_delete_files:
	    globus_l_replica_management_specify_action(arg_delete_files);
	    break;

	case arg_register:
	    globus_l_replica_management_specify_action(arg_register);
	    break;
	case arg_publish:
	    globus_l_replica_management_specify_action(arg_publish);
	    break;
	case arg_copy:
	    globus_l_replica_management_specify_action(arg_copy);
	    target_location_name = globus_libc_strdup(option->values[0]);
	    break;

	case arg_no_third_party:
	    third_party_transfer = GLOBUS_FALSE;
	    break;
	case arg_parallelism:
	    num_streams = atoi(option->values[0]);
	    break;
	case arg_tcp_buffer_size:
	    tcp_buffer_size = atoi(option->values[0]);
	    break;
	case arg_subject:
	    subject = globus_libc_strdup(option->values[0]);
	    break;
	case arg_dest_subject:
	    dest_subject = globus_libc_strdup(option->values[0]);
	    break;

	case arg_list_rollback_info:
	    globus_l_replica_management_specify_action(
		    arg_list_rollback_info);
	    replica_catalog_url = globus_libc_strdup(option->values[0]);
	    break;
	case arg_rollback:
	    globus_l_replica_management_specify_action(arg_rollback);
	    contact_string = globus_libc_strdup(option->values[0]);
	    break;
	case arg_restart:
	    globus_l_replica_management_specify_action(arg_restart);
	    contact_string = globus_libc_strdup(option->values[0]);
	    break;

	default:
	    globus_l_replica_management_arg_error(
		    "Error scanning arguments." );
	    break;
	}
    }

    globus_args_option_instance_list_free(&options_found);

    /*
     * check arguments
     */

    if (action == 0)
    {
	globus_l_replica_management_arg_error("Specify an action.");
    }

    if ( (object == 0)
	    && (action != arg_list_rollback_info)
	    && (action != arg_rollback) && (action != arg_restart) )
    {
	globus_l_replica_management_arg_error("Specify an object.");
    }

    if ( (collection_url == GLOBUS_NULL)
	    && (action != arg_list_rollback_info)
	    && (action != arg_rollback) && (action != arg_restart) )
    {
	if ( ( collection_url = globus_libc_getenv(
		"GLOBUS_REPLICA_CATALOG_COLLECTION") ) == GLOBUS_NULL )
	{
	    globus_l_replica_management_arg_error("Specify a collection.");
	}
    }

    if ((input_file_name != GLOBUS_NULL) && (location_name == GLOBUS_NULL))
    {
	globus_l_replica_management_arg_error(
		"Specify a location for files." );
    }

    if ( (object == arg_collection)
	    && (action != arg_create) && (action != arg_delete) )
    {
	globus_l_replica_management_arg_error(
		"Specify a corresponding action for collection." );
    }

    if ( (object == arg_location)
	    && (action != arg_create) && (action != arg_delete) )
    {
	globus_l_replica_management_arg_error(
		"Specify an appropriate action for location." );
    }

    if ( (object == arg_files)
	    && (action != arg_register) && (action != arg_publish)
	    && (action != arg_copy) && (action != arg_delete)
	    && (action != arg_delete_files) )
    {
	globus_l_replica_management_arg_error(
		"Specify an appropriate action for files." );
    }

    if ((object == arg_location) && (action == arg_create))
    {
	if (argc > 1)
	{
	    root_url = globus_libc_strdup(argv[1]);
	}
	else
	{
	    globus_l_replica_management_arg_error(
		    "Specify a root URL for this action." );
	}
    }

    if ((manager == GLOBUS_NULL) && (action != arg_list_rollback_info))
    {
	if ( ( manager = globus_libc_getenv(
		"GLOBUS_REPLICA_CATALOG_MANAGER" ) ) == GLOBUS_NULL )
	{
	    globus_l_replica_management_arg_error(
		    "Specify a manager with this action." );
	}
    }

    if (manager != GLOBUS_NULL)
    {
	if (password_file_name == GLOBUS_NULL)
	{
	    password = getpass("Password: ");
	}
	else
	{
	    if (strcmp(password_file_name,"-") == 0)
	    {
		fd = stdin;
	    }
	    else if ((fd = fopen(password_file_name, "r")) == NULL)
	    {
		globus_l_replica_management_arg_error(
			"Error opening password file for read." );
	    }
	    password = (char *) globus_libc_malloc(
		    GLOBUS_REPLICA_MANAGEMENT_BUFSIZE * sizeof(char) );
	    fgets(password, GLOBUS_REPLICA_MANAGEMENT_BUFSIZE, fd);
	    password[strlen(password)-1] = '\0';
	    if (strcmp(password_file_name,"-") != 0)
	    {
		fclose(fd);
	    }
	}
    }

    if (input_file_name != GLOBUS_NULL)
    {
	if (strcmp(input_file_name,"-") == 0)
	{
	    fd = stdin;
	}
	else if ((fd = fopen(input_file_name, "r")) == NULL)
	{
	    globus_l_replica_management_arg_error(
		    "Error opening input file for read." );
	}
	logical_file_names = (char **) globus_libc_malloc(
		GLOBUS_REPLICA_MANAGEMENT_BUFSIZE * sizeof(char *) );
	if (action == arg_publish)
	{
	    source_urls = (char **) globus_libc_malloc(
		    GLOBUS_REPLICA_MANAGEMENT_BUFSIZE * sizeof(char *) );
	}
	i = 0;
        
	while (1 != 2)
	{
	    logical_file_names[i] = (char *) globus_libc_malloc(
		    GLOBUS_REPLICA_MANAGEMENT_BUFSIZE * sizeof(char) );
	    if (action == arg_publish)
	    {
		source_urls[i] = (char *) globus_libc_malloc(
			GLOBUS_REPLICA_MANAGEMENT_BUFSIZE * sizeof(char) );
	    }
	    if (action == arg_publish)
	    {
		if ( fscanf( fd, "%s %s", logical_file_names[i],
			source_urls[i] ) == EOF )
		{
		    break;
		}
	    }
	    else
	    {
		if (fscanf(fd, "%s", logical_file_names[i]) == EOF)
		{
		    break;
		}
	    }
	    i++;
	    if ((i % GLOBUS_REPLICA_MANAGEMENT_BUFSIZE) == 0)
	    {
		logical_file_names = (char **) globus_libc_realloc(
			logical_file_names,
			(i + GLOBUS_REPLICA_MANAGEMENT_BUFSIZE)
				* sizeof(char *) );
		if (action == arg_publish)
		{
		    source_urls = (char **) globus_libc_realloc(
			    source_urls,
			    (i + GLOBUS_REPLICA_MANAGEMENT_BUFSIZE)
				    * sizeof(char *) );
		}
	    }
	}

        globus_libc_free(logical_file_names[i]);

        logical_file_names[i] = GLOBUS_NULL;

        if (action == arg_publish)
	{
            globus_libc_free(source_urls[i]);
	    source_urls[i] = GLOBUS_NULL;
	}
	if (strcmp(input_file_name,"-") != 0)
	{
	    fclose(fd);
	}
    }

    /* TODO: add checking for extraneous arguments */
    /* TODO: handle non-cleartext replica catalog authentication */

    /*
     * set up
     */

    globus_replica_management_handleattr_init(&handleattrs);

    /* XXX: should this be called on the handleattrs? */
    globus_replica_management_handleattr_set_third_party_transfer(
	    &handleattrs,
	    third_party_transfer );

    globus_replica_management_handle_init(&handle, &handleattrs);

    globus_replica_catalog_collection_handleattr_init(&collection_attrs);
    globus_replica_catalog_collection_handleattr_set_authentication_mode(
	    &collection_attrs,
	    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT,
	    manager,
	    password );

    globus_ftp_client_operationattr_init(&source_operationattrs);
    globus_ftp_client_operationattr_init(&dest_operationattrs);

    if (num_streams > 0)
    {
	globus_ftp_client_operationattr_set_mode(
		&source_operationattrs,
		GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
	globus_ftp_client_operationattr_set_mode(
		&dest_operationattrs,
		GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);

	parallelism.mode = GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
	parallelism.fixed.size = num_streams;
	globus_ftp_client_operationattr_set_parallelism(
		&source_operationattrs,
		&parallelism);
	globus_ftp_client_operationattr_set_parallelism(
		&dest_operationattrs,
		&parallelism);
    }

    if (tcp_buffer_size > 0)
    {
	tcp_buffer.mode = GLOBUS_FTP_CONTROL_TCPBUFFER_FIXED;
	tcp_buffer.fixed.size = tcp_buffer_size;
	globus_ftp_client_operationattr_set_tcp_buffer(
		&source_operationattrs,
		&tcp_buffer);
	globus_ftp_client_operationattr_set_tcp_buffer(
		&dest_operationattrs,
		&tcp_buffer);
    }

    if (subject != GLOBUS_NULL)
    {
	/* TODO: check if username and password have to be set in these
	 * attrs if specified in the url that is subsequently available
	 * in g_l_r_m_file_move */
	globus_ftp_client_operationattr_set_authorization(
		&source_operationattrs,
		GSS_C_NO_CREDENTIAL,
		NULL,
		NULL,
		NULL,
		subject);
    }

    if (dest_subject == GLOBUS_NULL)
    {
	dest_subject = subject;
    }

    if (dest_subject != GLOBUS_NULL)
    {
	/* TODO: check if username and password have to be set in these
	 * attrs if specified in the url that is subsequently available
	 * in g_l_r_m_file_move */
	globus_ftp_client_operationattr_set_authorization(
		&dest_operationattrs,
		GSS_C_NO_CREDENTIAL,
		NULL,
		NULL,
		NULL,
		dest_subject);
    }

    /*
     * do actions
     */

    switch (object)
    {
    case arg_collection:
	switch (action)
	{
	case arg_create:
	    err = globus_replica_management_collection_create(
		    &handle,
		    collection_url,
		    &collection_attrs );
	    globus_l_replica_management_check_error(err);
	    break;
	case arg_delete:
	    err = globus_replica_management_collection_delete(
		    &handle,
		    collection_url,
		    &collection_attrs );
	    globus_l_replica_management_check_error(err);
	    break;
	default:
	    globus_l_replica_management_arg_error(
		    "Error acting on collection arguments." );
	    break;
	}
	break;

    case arg_location:
	switch (action)
	{
	case arg_create:
	    err = globus_replica_management_location_create(
		    &handle,
		    collection_url,
		    &collection_attrs,
		    location_name,
		    &source_operationattrs,
		    root_url );
	    globus_l_replica_management_check_error(err);
	    break;
	case arg_delete:
	    err = globus_replica_management_location_delete(
		    &handle,
		    collection_url,
		    &collection_attrs,
		    location_name );
	    globus_l_replica_management_check_error(err);
	    break;
	default:
	    globus_l_replica_management_arg_error(
		    "Error acting on location arguments." );
	    break;
	}
	break;

    case arg_files:
	switch (action)
	{
	case arg_register:
	    for (i = 0; logical_file_names[i];i++)
	    {
		err = globus_replica_management_file_register(
			&handle,
			collection_url,
			&collection_attrs,
			location_name,
			&dest_operationattrs,
			logical_file_names[i] );
		globus_l_replica_management_check_error(err);
	    }
	    break;
	case arg_publish:
	    for (i = 0; logical_file_names[i];i++)
	    {
		err = globus_replica_management_file_publish(
			&handle,
			source_urls[i],
			&source_operationattrs,
			collection_url,
			&collection_attrs,
			location_name,
			&dest_operationattrs,
			logical_file_names[i],
			/* TODO: allow for a commandline param */
			GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW,
			GLOBUS_NULL,
			GLOBUS_NULL );
		globus_l_replica_management_check_error(err);
	    }
	    break;
	case arg_copy:
	    for (i = 0; logical_file_names[i];i++)
	    {
		err = globus_replica_management_file_copy(
			&handle,
			collection_url,
			&collection_attrs,
			location_name,
			&source_operationattrs,
			target_location_name,
			&dest_operationattrs,
			logical_file_names[i],
			/* TODO: allow for a commandline param */
			GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW,
			GLOBUS_NULL,
			GLOBUS_NULL );
		globus_l_replica_management_check_error(err);
	    }
	    break;
	case arg_delete:
	    for (i = 0; logical_file_names[i];i++)
	    {
		err = globus_replica_management_file_delete(
			&handle,
			collection_url,
			&collection_attrs,
			location_name,
			&dest_operationattrs,
			logical_file_names[i],
			GLOBUS_FALSE );
		globus_l_replica_management_check_error(err);
	    }
	    break;
	case arg_delete_files:
	    for (i = 0; logical_file_names[i];i++)
	    {
		err = globus_replica_management_file_delete(
			&handle,
			collection_url,
			&collection_attrs,
			location_name,
			&dest_operationattrs,
			logical_file_names[i],
			GLOBUS_TRUE );
		globus_l_replica_management_check_error(err);
	    }
	    break;
	default:
	    globus_l_replica_management_arg_error(
		    "Error acting on file arguments." );
	    break;
	}
	break;

    default:
	switch (action)
	{
	case arg_list_rollback_info:
	    globus_replica_catalog_entry_set_init(&entries);
	    err = globus_replica_catalog_list_rollback_entries(
		    replica_catalog_url,
		    GLOBUS_NULL,
		    GLOBUS_NULL,
		    &entries );
	    globus_l_replica_management_check_error(err);
	    globus_l_replica_catalog_print_entries(entries);
	    globus_replica_catalog_entry_set_destroy(&entries);
	    break;
	case arg_rollback:
	    err = globus_replica_management_rollback(
		    &handle,
		    contact_string,
		    &collection_attrs,
		    &source_operationattrs );
	    globus_l_replica_management_check_error(err);
	    break;
	case arg_restart:
	    err = globus_replica_management_restart(
		    &handle,
		    contact_string,
		    &collection_attrs,
		    &source_operationattrs,
		    &dest_operationattrs,
		    GLOBUS_NULL,
		    GLOBUS_NULL );
	    globus_l_replica_management_check_error(err);
	    break;
	default:
	    globus_l_replica_management_arg_error(
		    "Error acting on arguments." );
	    break;
	}
	break;
    }

    /*
     * clean up
     */

cleanup:
    globus_ftp_client_operationattr_destroy(&source_operationattrs);
    globus_ftp_client_operationattr_destroy(&dest_operationattrs);
    globus_replica_management_handleattr_destroy(&handleattrs);
    globus_replica_management_handle_destroy(&handle);
    globus_replica_catalog_collection_handleattr_destroy(&collection_attrs);

    globus_module_deactivate_all();

    exit(errors);
}

